#include <JUI/Widget.hpp>


namespace JUI {

    Widget::Widget() {
        name = "Widget";
        children = std::vector<Widget *>();
    }

    Widget::Widget(Widget* parent) : Widget()
    {
         this->SetParent(parent);
    }

    void Widget::SetParent(Widget* parent) {
        // hold a reference to this so it doesn't get collected as we're working.
        Widget* oldParent = this->parent;

        // New parent is old parent, do nothing, maybe raise a warning.
        if (parent == oldParent)
            return;

        // Don't allow for an instance to be parented to itself
        if (this == parent)
            throw std::runtime_error("Cannot parent a widget to itself.");
        // You're trying to break the linearity of the object hierarchy by creating a circle in the graph.
        if (this->IsAncestorOf(parent))
            throw std::runtime_error("Cannot create circular reference.");

        for (Widget* ancestor: this->GetAncestors()) {
            if (oldParent && !ancestor->IsAncestorOf(parent) && parent != ancestor) {
                ancestor->DescendantRemoved(this);
                for (Widget* descendant : this->GetDescendants()) {
                    ancestor->DescendantRemoved(descendant);
                }
            }
        }

        // Remove ourselves from our parent (if we have one)
        if (this->parent) {
            // this->parent->ChildRemoved(this)
            std::erase(parent->children, this);
        }

        // Update our old parent to the new one
        this->parent = parent;

        // If our parent is set to nullptr, we can't update it's vector of children
        if (!parent) return;

        // Add ourselves to our new parent's list of children
        parent->children.emplace_back(this);
        //newParent->ChildAdded(this);

        for (Widget* ancestor : this->GetAncestors()) {
            if (!oldParent || (!oldParent->IsDescendantOf(parent) && oldParent != ancestor)) {
                // Don't fire unless an instance is actually a new descendant
                ancestor->DescendantAdded(this);
                for (Widget* descendant: this->GetDescendants()) {
                    ancestor->DescendantAdded(descendant);
                }
            }
        }
    }

    bool Widget::IsDescendantOf(Widget* ancestor) {
        if (ancestor == nullptr) return false;

        Widget* instance = this;
        while (instance->GetParent()) {
            instance = instance->GetParent();
            if (instance == ancestor)
                return true;
        }
        return false;
    }

    bool Widget::IsAncestorOf(Widget* descendant) {
        if (descendant == nullptr) return false;
        Widget* instance = descendant;
        while (instance->GetParent()) {
            instance = instance->GetParent();
            if (instance == this)
                return true;
        }
        return false;
    }

    Vector2 Widget::GetAnchorPoint() const {
        return anchor_point;
    }

    UDim Widget::GetPaddingLeft() const { return pad_left;}
    UDim Widget::GetPaddingTop() const { return pad_top;}
    UDim Widget::GetPaddingRight() const { return pad_right;}
    UDim Widget::GetPaddingBottom() const { return pad_bottom;}

    void Widget::SetPaddingLeft  (const UDim& pl) { pad_left   = pl;}
    void Widget::SetPaddingTop   (const UDim& pt) { pad_top    = pt;}
    void Widget::SetPaddingRight (const UDim& pr) { pad_right  = pr;}
    void Widget::SetPaddingBottom(const UDim& pb) { pad_bottom = pb;}

    UDim Widget::GetMarginLeft  () const { return margin_left;}
    UDim Widget::GetMarginTop   () const { return margin_top;}
    UDim Widget::GetMarginRight () const { return margin_right;}
    UDim Widget::GetMarginBottom() const { return margin_bottom;}

    void Widget::SetMarginLeft(const UDim& ml) { margin_left = ml;}
    void Widget::SetMarginTop(const UDim& mt) { margin_top = mt;}
    void Widget::SetMarginRight(const UDim& mr) { margin_right = mr;}
    void Widget::SetMarginBottom(const UDim& mb) { margin_bottom = mb;}

    std::string Widget::GetName() const { return name; }
    void Widget::SetName(const std::string& new_name) { name = new_name;}

    bool Widget::IsVisible() const { return visible; }
    void Widget::SetVisible(bool enabled) { visible = enabled;}

    Widget* Widget::GetParent() const { return parent;}

    UDim2 Widget::GetPosition() const { return position; }
    void Widget::SetPosition(const UDim2& pos) { position = pos; }

    UDim2 Widget::GetSize() const { return size;}
    void Widget::SetSize(const UDim2& s) { size = s; }

    float Widget::GetAbsoluteRotation() const {
        // TODO: implement correctly!!
        return rotation;
    }


    std::vector<Widget *> Widget::GetChildren() { return children; }

    std::vector<Widget *> Widget::GetDescendants() {
        std::vector<Widget*> descendants;
        for (auto& child: this->GetChildren()) {
            descendants.push_back(child);
            std::vector<Widget*>  recursiveDescendants = child->GetDescendants();
            descendants.insert(descendants.end(), recursiveDescendants.begin(), recursiveDescendants.end());
        }
        return descendants;
    }

    std::vector<Widget*> Widget::GetAncestors() {
        std::vector<Widget*> ancestors;
        for (Widget *ancestor = this->parent; ancestor; ancestor = ancestor->parent) {
            ancestors.push_back(ancestor);
        }
        return ancestors;
    }



    Vector2 Widget::GetAbsolutePosition() const {
        auto child_pos_scale = this->GetPosition().GetScale();
        auto child_pos_pixels = this->GetPosition().GetPixels();
        auto parent_abs_size = this->GetParent()->GetAbsoluteSize();
        auto parent_abs_pos = this->GetParent()->GetAbsolutePosition();

        auto abs_size = GetAbsoluteSize();

        auto anchor_offset = abs_size * GetAnchorPoint();

        UDim padding_h = parent->GetPaddingLeft(); // TODO: Factor in PaddingRight
        UDim padding_v = parent->GetPaddingTop();  // TODO: Factor in PaddingBottom

        float calculated_padding_x = padding_h.Pixels + (padding_h.Scale * parent_abs_size.x);
        float calculated_padding_y = padding_v.Pixels + (padding_v.Scale * parent_abs_size.y);

        Vector2 padding_offset = {calculated_padding_x, calculated_padding_y};


        Vector2 absolute_position =
            parent_abs_pos + child_pos_pixels + (parent_abs_size * child_pos_scale) + padding_offset;

        return absolute_position - anchor_offset;
    }

    // TODO: implement padding shrinking abs_size
    Vector2 Widget::GetAbsoluteSize() const {
        Vector2 child_size_scale = this->GetSize().GetScale();
        Vector2 child_size_px = this->GetSize().GetPixels();

        Vector2 parent_abs_size = this->GetParent()->GetAbsoluteSize();

        UDim padding_h = parent->GetPaddingLeft() + parent->GetPaddingRight();
        float final_pad_x = padding_h.Pixels + (padding_h.Scale * parent->GetAbsoluteSize().x);

        UDim padding_v = parent->GetPaddingTop() + parent->GetPaddingBottom();
        float final_pad_y = padding_v.Pixels + (padding_v.Scale * parent->GetAbsoluteSize().y);

        Vector2 pad_size_reduction = {final_pad_x, final_pad_y};

        Vector2 abs_size = child_size_px + (parent_abs_size * child_size_scale) - pad_size_reduction;

        // TODO: Take into account constraints on widgets.
        return abs_size;
    }


    // TODO: Consider std::optional
    Widget* Widget::FindFirstChild(const std::string& search_name) {
        for (auto& child : children) {
            if (child->GetName() == search_name)
                return child;
        }
        return nullptr;
    }


    void Widget::Draw() {
        if (!visible)
            return;

        DrawChildWidgets();
    }

    void Widget::Update(float delta) {

        UpdateChildWidgets(delta);
    }

    void Widget::DrawChildWidgets() {
        for (auto child : children) {
            child->Draw();
        }
    }

    void Widget::UpdateChildWidgets(float delta) {
        for (auto child : children) {
            child->Update(delta);
        }
    }

    Widget* Widget::GetFamilyTreeRoot() const {
        // This is stupid, Fix ASAP!
        auto parent = this->GetParent();
        if (parent->GetParent() == nullptr)
            return parent;
        return parent->GetFamilyTreeRoot();
    }


}